import pytest

from . import format_smart80, format_smart80_align_dynamic, format_smart80_align_fixed32


@pytest.mark.golden_test("golden/smart80/default/*.yml")
def test_smart80_idempotent(golden):
    assert golden["input"] is not None
    output = format_smart80(golden["output"])
    assert (
        output == golden.out["output"]
    ), "Formatter with default options is not idempotent"


@pytest.mark.golden_test("golden/smart80/align/dynamic/*.yml")
def test_smart80_align_dynamic_idempotent(golden):
    assert golden["input"] is not None
    output = format_smart80_align_dynamic(golden["output"])
    assert (
        output == golden.out["output"]
    ), "Formatter with dynamic alignment is not idempotent"


@pytest.mark.golden_test("golden/smart80/align/fixed32/*.yml")
def test_smart80_align_fixed32_idempotent(golden):
    assert golden["input"] is not None
    output = format_smart80_align_fixed32(golden["output"])
    assert (
        output == golden.out["output"]
    ), "Formatter with fixed alignment is not idempotent"



import pytest

from . import format_simple, format_simple_align_dynamic, format_simple_align_fixed32


@pytest.mark.golden_test("golden/simple/default/*.yml")
def test_simple_idempotent(golden):
    assert golden["input"] is not None
    output = format_simple(golden["output"])
    assert (
        output == golden.out["output"]
    ), "Formatter with default options is not idempotent"


@pytest.mark.golden_test("golden/simple/align/dynamic/*.yml")
def test_simple_align_dynamic_idempotent(golden):
    assert golden["input"] is not None
    output = format_simple_align_dynamic(golden["output"])
    assert (
        output == golden.out["output"]
    ), "Formatter with dynamic alignment is not idempotent"


@pytest.mark.golden_test("golden/simple/align/fixed32/*.yml")
def test_simple_align_fixed32_idempotent(golden):
    assert golden["input"] is not None
    output = format_simple_align_fixed32(golden["output"])
    assert (
        output == golden.out["output"]
    ), "Formatter with fixed alignment is not idempotent"



import pytest

from . import format_smart1k, format_smart1k_align_dynamic, format_smart1k_align_fixed32


@pytest.mark.golden_test("golden/smart1k/default/*.yml")
def test_smart1k(golden):
    output = format_smart1k(golden["input"])
    assert output == golden.out["output"]


@pytest.mark.golden_test("golden/smart1k/align/dynamic/*.yml")
def test_smart1k_align_dynamic(golden):
    output = format_smart1k_align_dynamic(golden["input"])
    assert output == golden.out["output"]


@pytest.mark.golden_test("golden/smart1k/align/fixed32/*.yml")
def test_smart1k_align_fixed32(golden):
    output = format_smart1k_align_fixed32(golden["input"])
    assert output == golden.out["output"]



import io
import sys
import tokenize
from pathlib import Path
from typing import Optional, Union

import click
import tree_sitter_talon
from doc_printer import DocRenderer, SimpleDocRenderer, SimpleLayout, SmartDocRenderer

from talonfmt.formatter import ParseError, TalonFormatter


def talonfmt(
    contents: str,
    *,
    encoding: str = "utf-8",
    indent_size: int = 4,
    max_line_width: Optional[int] = None,
    align_match_context: bool = False,
    align_match_context_at: Optional[int] = None,
    align_short_commands: bool = False,
    align_short_commands_at: Optional[int] = None,
    format_comments: bool = True,
) -> str:
    # Enable align_match_context if align_match_context_at is set:
    merged_match_context: Union[bool, int]
    if isinstance(align_match_context_at, int):
        merged_match_context = align_match_context_at
    else:
        merged_match_context = align_match_context

    # Enable align_short_commands if align_short_commands_at is set:
    merged_short_commands: Union[bool, int]
    if isinstance(align_short_commands_at, int):
        merged_short_commands = align_short_commands_at
    else:
        merged_short_commands = align_short_commands

    # Create an instance of TalonFormatter
    talon_formatter = TalonFormatter(
        indent_size=indent_size,
        align_match_context=merged_match_context,
        align_short_commands=merged_short_commands,
        format_comments=format_comments,
    )

    # Create an instance of DocRenderer
    doc_renderer: DocRenderer
    if max_line_width:
        doc_renderer = SmartDocRenderer(max_line_width=max_line_width)
    else:
        simple_layout: SimpleLayout
        if align_match_context is not False or align_short_commands is not False:
            simple_layout = SimpleLayout.LongestLines
        else:
            simple_layout = SimpleLayout.ShortestLines
        doc_renderer = SimpleDocRenderer(simple_layout=simple_layout)

    ast = tree_sitter_talon.parse(contents, encoding=encoding)
    doc = talon_formatter.format(ast)
    return doc_renderer.to_str(doc)


@click.command(name="talonfmt")
@click.argument(
    "path",
    nargs=-1,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=True, readable=True, path_type=Path
    ),
)
@click.option("--indent-size", type=int, default=4, show_default=True)
@click.option("--max-line-width", type=int, show_default=True)
@click.option(
    "--align-match-context/--no-align-match-context",
    default=False,
    show_default=True,
)
@click.option(
    "--align-match-context-at",
    type=int,
)
@click.option(
    "--align-short-commands/--no-align-short-commands",
    default=False,
    show_default=True,
)
@click.option(
    "--align-short-commands-at",
    type=int,
)
@click.option(
    "--format-comments/--no-format-comments",
    default=True,
    show_default=True,
)
@click.option(
    "--in-place/--stdout",
    default=False,
    show_default=True,
)
@click.option(
    "--fail-on-change/--no-fail-on-change",
    default=False,
    show_default=True,
)
@click.option(
    "--fail-on-error/--no-fail-on-error",
    default=False,
    show_default=True,
)
@click.option(
    "--verbose/--quiet",
    default=True,
    show_default=True,
)
def cli(
    *,
    path: tuple[Path, ...],
    indent_size: int,
    max_line_width: Optional[int],
    align_match_context: bool,
    align_match_context_at: Optional[int],
    align_short_commands: bool,
    align_short_commands_at: Optional[int],
    format_comments: bool,
    in_place: bool,
    fail_on_change: bool,
    fail_on_error: bool,
    verbose: bool,
):
    files_changed: list[str] = []

    def readfile(filename: Path) -> tuple[str, str]:
        with filename.open(mode="rb") as fp:
            bytes_on_disk = fp.read()
        encoding, _ = tokenize.detect_encoding(io.BytesIO(bytes_on_disk).readline)
        with io.TextIOWrapper(io.BytesIO(bytes_on_disk), encoding) as wrapper:
            contents = wrapper.read()
        return (contents, encoding)

    def format(
        contents: str, *, encoding: str, filename: Optional[str] = None
    ) -> Optional[str]:
        try:
            if verbose:
                sys.stderr.write(f"Formatting {filename}...\n")
            output = talonfmt(
                contents,
                encoding=encoding,
                indent_size=indent_size,
                max_line_width=max_line_width,
                align_match_context=align_match_context,
                align_match_context_at=align_match_context_at,
                align_short_commands=align_short_commands,
                align_short_commands_at=align_short_commands_at,
                format_comments=format_comments,
            )
            if contents != output and filename:
                if fail_on_change and verbose:
                    sys.stderr.write(f"File changed!")
                files_changed.append(filename)
            return output
        except ParseError as e:
            sys.stderr.write(e.message(contents=contents, filename=filename))
            if fail_on_error:
                exit(1)
        return None

    def format_file(filename: Path):
        contents, encoding = readfile(filename)
        output = format(contents, encoding=encoding, filename=str(filename))
        if output:
            if in_place:
                with filename.open(mode="w") as handle:
                    handle.write(output)
            else:
                sys.stdout.write(output)

    if path:
        for file_or_dir in path:
            if file_or_dir.is_file():
                format_file(file_or_dir)
            if file_or_dir.is_dir():
                for file in file_or_dir.glob("**/*.talon"):
                    format_file(file)
    else:
        contents = "\n".join(sys.stdin.readlines())
        encoding = sys.stdin.encoding
        output = format(contents, encoding=encoding)
        if output:
            sys.stdout.write(output)

    if fail_on_change and files_changed:
        exit(2)
    else:
        exit(0)


def main():
    cli(prog_name="talonfmt")


if __name__ == "__main__":
    main()



import pytest

from . import format_smart80, format_smart80_align_dynamic, format_smart80_align_fixed32


@pytest.mark.golden_test("golden/smart80/default/*.yml")
def test_smart80(golden):
    output = format_smart80(golden["input"])
    assert output == golden.out["output"]


@pytest.mark.golden_test("golden/smart80/align/dynamic/*.yml")
def test_smart80_align_dynamic(golden):
    output = format_smart80_align_dynamic(golden["input"])
    assert output == golden.out["output"]


@pytest.mark.golden_test("golden/smart80/align/fixed32/*.yml")
def test_smart80_align_fixed32(golden):
    output = format_smart80_align_fixed32(golden["input"])
    assert output == golden.out["output"]



import pytest

from . import format_simple, format_simple_align_dynamic, format_simple_align_fixed32


@pytest.mark.golden_test("golden/simple/default/*.yml")
def test_simple(golden):
    output = format_simple(golden["input"])
    assert output == golden.out["output"]


@pytest.mark.golden_test("golden/simple/align/dynamic/*.yml")
def test_simple_align_dynamic(golden):
    output = format_simple_align_dynamic(golden["input"])
    assert output == golden.out["output"]


@pytest.mark.golden_test("golden/simple/align/fixed32/*.yml")
def test_simple_align_fixed32(golden):
    output = format_simple_align_fixed32(golden["input"])
    assert output == golden.out["output"]



import dataclasses
from collections.abc import Iterator
from typing import Optional, Sequence

from tree_sitter_talon import Point, TalonError


@dataclasses.dataclass
class ParseError(Exception):
    start_position: Point
    end_position: Point

    def __init__(self, talon_error: TalonError):
        self.start_position = talon_error.start_position
        self.end_position = talon_error.end_position

    @staticmethod
    def point_to_str(point: Point) -> str:
        return f"line {point.row}, column {point.column}"

    def range(self) -> str:
        if self.start_position.row == self.end_position.row:
            return f"on line {self.start_position.row} between column {self.start_position.column} and {self.end_position.column}"
        else:
            return f" between {self.point_to_str(self.start_position)} and {self.point_to_str(self.end_position)}"

    def annotated_region(self, contents: str) -> str:
        def annotated_lines(lines: Sequence[str]) -> Iterator[str]:
            for l, line in enumerate(lines):
                yield line
                is_first_line = l == 0
                is_last_line = l == len(lines) - 1
                start = self.start_position.column if is_first_line else 0
                end = self.end_position.column if is_last_line else len(line)
                annotation: list[str] = []
                for c, _ in enumerate(line):
                    if c <= start or end < c:
                        annotation.append(" ")
                    else:
                        annotation.append("^")
                yield "".join(annotation)

        lines = contents.splitlines()
        lines = lines[self.start_position.row : self.end_position.row + 1]
        return "\n".join(annotated_lines(lines))

    def message(self, *, contents: str, filename: Optional[str] = None) -> str:
        return "".join(
            (
                f"Parse error ",
                f"in {filename} " if filename else "",
                f"{self.range()}:\n",
                f"{self.annotated_region(contents)}\n",
            )
        )



import dataclasses
import itertools
from collections.abc import Iterable, Iterator
from functools import singledispatchmethod
from typing import Optional, TypeVar, Union

from doc_printer import (
    Doc,
    Empty,
    Fail,
    Line,
    Map,
    Space,
    Table,
    Text,
    alt,
    angles,
    braces,
    brackets,
    cat,
    create_table,
    double_quote,
    group,
    nest,
    parens,
    row,
)
from tree_sitter_talon import (
    Node,
    TalonAction,
    TalonAnd,
    TalonArgumentList,
    TalonAssignment,
    TalonBinaryOperator,
    TalonBlock,
    TalonCapture,
    TalonChoice,
    TalonCommand,
    TalonComment,
    TalonContext,
    TalonDocstring,
    TalonEndAnchor,
    TalonError,
    TalonExpression,
    TalonFloat,
    TalonIdentifier,
    TalonImplicitString,
    TalonIncludeTag,
    TalonInteger,
    TalonInterpolation,
    TalonKeyAction,
    TalonList,
    TalonMatch,
    TalonNot,
    TalonNumber,
    TalonOperator,
    TalonOptional,
    TalonOr,
    TalonParenthesizedExpression,
    TalonParenthesizedRule,
    TalonRegexEscapeSequence,
    TalonRepeat,
    TalonRepeat1,
    TalonRule,
    TalonSeq,
    TalonSettings,
    TalonSleepAction,
    TalonSourceFile,
    TalonStartAnchor,
    TalonString,
    TalonStringContent,
    TalonStringEscapeSequence,
    TalonVariable,
    TalonWord,
)

from .parse_error import ParseError

TalonBlockLevelMatch = Union[
    TalonAnd,
    TalonNot,
    TalonMatch,
    TalonOr,
]

TalonBlockLevel = Union[
    TalonSourceFile,
    TalonContext,
    TalonIncludeTag,
    TalonSettings,
    TalonCommand,
    TalonBlock,
    TalonAssignment,
    TalonExpression,
    TalonComment,
    TalonDocstring,
]

NodeVar = TypeVar("NodeVar", bound=Node)


def is_short_command(node: TalonCommand) -> bool:
    return len(node.children) + len(node.script.children) == 1


def block_with_comments(
    comments: Iterable[TalonComment], block: TalonBlock
) -> TalonBlock:
    return TalonBlock(
        text=block.text,
        type_name=block.type_name,
        start_position=block.start_position,
        end_position=block.end_position,
        children=[*comments, *block.children],
    )


@dataclasses.dataclass
class TalonFormatter:
    indent_size: int
    align_match_context: Union[bool, int]
    align_short_commands: Union[bool, int]
    format_comments: bool

    @singledispatchmethod
    def format(self, node: Node) -> Doc:
        """
        Format any node as a document.
        """
        # NOTE: these should implement format_lines
        if isinstance(
            node,
            (
                TalonSourceFile,
                TalonContext,
                TalonIncludeTag,
                TalonSettings,
                TalonCommand,
                TalonBlock,
                TalonAssignment,
                TalonExpression,
            ),
        ):
            return cat(self.format_lines(node))
        elif isinstance(node, TalonError):
            raise ParseError(node)
        else:
            raise TypeError(type(node))

    @singledispatchmethod
    def format_lines(self, node: TalonBlockLevel) -> Iterator[Doc]:
        """
        Format any block-level node as a series of lines.
        """
        if isinstance(node, (TalonComment, TalonDocstring)):
            yield self.format(node)
        elif isinstance(node, (TalonAnd, TalonNot, TalonMatch, TalonOr)):
            yield from self.format_lines_match(node, under_and=False, under_not=False)
        elif isinstance(node, TalonError):
            raise ParseError(node)
        else:
            raise TypeError(type(node))

    @singledispatchmethod
    def format_lines_match(
        self,
        match: TalonBlockLevelMatch,
        *,
        under_and: bool,
        under_not: bool,
    ) -> Iterator[Doc]:
        """
        Format any match statement or comment as a series of lines.
        """
        if isinstance(match, TalonError):
            raise ParseError(match)
        else:
            raise TypeError(type(match))

    def format_children(self, children: Iterable[Node]) -> Iterator[Doc]:
        for child in self.store_comments_with_type(children, node_type=Node):
            if isinstance(child, Iterable):
                yield from self.format_children(child)
            else:
                yield self.format(child)

    ###########################################################################
    # Format: Source Files
    ###########################################################################

    @format_lines.register
    def _(self, node: TalonSourceFile) -> Iterator[Doc]:

        # Used to emit the context header separator.
        in_header: bool = True

        # Used to buffer comments to ensure that they're split correctly
        # between the header and body.
        comment_buffer: list[Doc] = []

        def clear_comment_buffer() -> Iterator[Doc]:
            if comment_buffer:
                yield from comment_buffer
                comment_buffer.clear()

        if self.align_short_commands is True:
            # Used to buffer short commands to group them as tables.
            short_command_buffer: list[Doc] = []

            def clear_short_command_buffer() -> Iterator[Doc]:
                if short_command_buffer:
                    table = create_table(short_command_buffer)
                    if table:
                        yield alt(cat(short_command_buffer), table)
                    else:
                        yield from short_command_buffer
                    short_command_buffer.clear()

        # Used to insert blank lines.
        previous_line: int = 0

        for child in node.children:
            extra_blank_line: bool = child.start_position.row - previous_line >= 2

            if in_header and isinstance(child, TalonComment):
                # NOTE: buffer comments in context header
                if extra_blank_line:
                    comment_buffer.append(Line)
                comment_buffer.append(self.format(child))
            elif isinstance(child, TalonContext):
                # NOTE: format context header
                assert in_header
                yield from clear_comment_buffer()
                yield from self.format_lines(child)
            else:
                # NOTE: first body-only node, end the context header
                if in_header and isinstance(
                    child, (TalonIncludeTag, TalonSettings, TalonCommand)
                ):
                    yield Text("-") / Line
                    # NOTE: no blank lines after "-"
                    if comment_buffer:
                        yield from itertools.dropwhile(
                            lambda doc: doc is Line, clear_comment_buffer()
                        )
                    else:
                        extra_blank_line = False
                    in_header = False

                if self.align_short_commands is True:
                    if isinstance(child, TalonCommand) and is_short_command(child):
                        # NOTE: buffer short command
                        short_command_buffer.extend(self.format_lines(child))
                    else:
                        # NOTE: long command or other node, clear short command buffer
                        yield from clear_short_command_buffer()
                        if extra_blank_line:
                            yield Line
                        yield from self.format_lines(child)
                else:
                    if extra_blank_line:
                        yield Line
                    yield from self.format_lines(child)

            # Update previous line.
            previous_line = child.end_position.row

        # NOTE: no body-only node, end the context header
        if in_header:
            yield Text("-") / Line
            yield from clear_comment_buffer()

        # NOTE: clear remaining short commands in buffer
        if self.align_short_commands is True:
            yield from clear_short_command_buffer()

    ###########################################################################
    # Format: Context Header
    ###########################################################################

    @format_lines.register
    def _(self, node: TalonContext) -> Iterator[Doc]:
        for child in node.children:
            lines = list(self.format_lines(child))
            yield from self.get_formatted_comments()
            yield from lines

    @format_lines_match.register
    def _(self, match: TalonAnd, under_and: bool, under_not: bool) -> Iterator[Doc]:
        for child in match.children:
            if isinstance(child, TalonComment):
                yield from self.format_lines(child)
            else:
                yield from self.format_lines_match(child, under_and, under_not)
                under_and = True

    @format_lines_match.register
    def _(self, match: TalonNot, under_and: bool, under_not: bool) -> Iterator[Doc]:
        for child in match.children:
            if isinstance(child, TalonComment):
                yield from self.format_lines(child)
            else:
                yield from self.format_lines_match(child, under_and, under_not)
                under_not = True

    @format_lines_match.register
    def _(self, match: TalonOr, under_and: bool, under_not: bool) -> Iterator[Doc]:
        for child in match.children:
            if isinstance(child, TalonComment):
                yield from self.format_lines(child)
            else:
                lines = list(self.format_lines_match(child, under_and, under_not))
                yield from self.get_formatted_comments()
                yield from lines

    @format_lines_match.register
    def _(self, match: TalonMatch, under_and: bool, under_not: bool) -> Iterator[Doc]:
        self.assert_only_comments(match.children)
        match_keywords = self.format_match_keywords(under_and, under_not)
        key = cat(match_keywords, self.format(match.key))
        pattern = self.format(match.pattern)
        yield alt(self.format_match_alternatives(key, pattern))

    def format_match_keywords(self, under_and: bool, under_not: bool) -> Iterator[Doc]:
        if under_and:
            yield Text("and")
            yield Space
        if under_not:
            yield Text("not")
            yield Space

    def format_match_alternatives(self, key: Doc, pattern: Doc) -> Iterator[Doc]:
        # Standard
        yield key / ":" // pattern / Line

        # Aligned alternative
        if self.align_match_context:
            if self.align_match_context is True:
                yield row(
                    key / ":",
                    pattern,
                    table_type="match",
                )
            else:
                yield row(
                    key / ":",
                    pattern,
                    table_type="match",
                    min_col_widths=(self.align_match_context,),
                )

    ###########################################################################
    # Format: Tag Includes
    ###########################################################################

    @format_lines.register
    def _(self, node: TalonIncludeTag) -> Iterator[Doc]:
        self.assert_only_comments(node.children)
        yield from self.get_formatted_comments()
        yield "tag():" // self.format(node.tag) / Line

    ###########################################################################
    # Format: Settings
    ###########################################################################

    @format_lines.register
    def _(self, node: TalonSettings) -> Iterator[Doc]:
        block = self.get_node_with_type(node.children, node_type=TalonBlock)
        block = block_with_comments(self.get_comments(), block)
        yield "settings():" // nest(
            self.indent_size,
            Line,
            self.format(block),
        )

    ###########################################################################
    # Format: Commands
    ###########################################################################

    @format_lines.register
    def _(self, node: TalonCommand) -> Iterator[Doc]:
        rule = self.format(node.rule)
        yield from self.get_formatted_comments()

        # Merge comments on this node into the block node.
        block = block_with_comments(node.children, node.script)
        script = self.format(block)

        # (1): a line-break after the rule, e.g.,
        #
        # select camel left:
        #     user.extend_camel_left()
        #
        alt1 = cat(
            rule / ":",
            nest(self.indent_size, Line, script),
        )

        # (2): the rule and a single-line talon script on the same line, e.g.,
        #
        # select camel left: user.extend_camel_left()
        #
        if len(block.children) == 1:
            alt2 = self.format_short_command(rule, script)
        else:
            alt2 = Fail

        yield alt1 | alt2

    def format_short_command(self, rule: Doc, script: Doc) -> Doc:
        if self.align_short_commands:
            if self.align_short_commands is True:
                return row(
                    rule / ":",
                    group(script),
                    table_type="command",
                )
            else:
                return row(
                    rule / ":",
                    group(script),
                    table_type="command",
                    min_col_widths=(self.align_short_commands,),
                )
        else:
            return rule / ":" // group(script) / Line

    ###########################################################################
    # Format: Statements
    ###########################################################################

    @format_lines.register
    def _(self, node: TalonBlock) -> Iterator[Doc]:
        for child in node.children:
            for line in self.format_lines(child):
                yield from self.get_formatted_comments()
                yield line

    @format_lines.register
    def _(self, node: TalonAssignment) -> Iterator[Doc]:
        self.assert_only_comments(node.children)
        yield self.format(node.left) // "=" // self.format(node.right) / Line

    @format_lines.register
    def _(self, node: TalonExpression) -> Iterator[Doc]:
        self.assert_only_comments(node.children)
        yield self.format(node.expression) / Line

    ###########################################################################
    # Format: Expressions
    ###########################################################################

    @format.register
    def _(self, node: TalonAction) -> Doc:
        self.assert_only_comments(node.children)
        return self.format(node.action_name) / parens(self.format(node.arguments))

    @format.register
    def _(self, node: TalonArgumentList) -> Doc:
        return ("," / Space).join(self.format_children(node.children))

    @format.register
    def _(self, node: TalonBinaryOperator) -> Doc:
        self.assert_only_comments(node.children)
        return (
            self.format(node.left)
            // self.format(node.operator)
            // self.format(node.right)
        )

    @format.register
    def _(self, node: TalonIdentifier) -> Doc:
        return Text.words(node.text, collapse_whitespace=True)

    @format.register
    def _(self, node: TalonKeyAction) -> Doc:
        self.assert_only_comments(node.children)
        return "key" / parens(self.format(node.arguments))

    @format.register
    def _(self, node: TalonOperator) -> Doc:
        return Text.words(node.text, collapse_whitespace=True)

    @format.register
    def _(self, node: TalonParenthesizedExpression) -> Doc:
        return parens(self.format(self.get_node(node.children)))

    @format.register
    def _(self, node: TalonRegexEscapeSequence) -> Doc:
        if node.children:
            return braces(self.format_children(node.children))
        else:
            return braces(Empty)

    @format.register
    def _(self, node: TalonSleepAction) -> Doc:
        self.assert_only_comments(node.children)
        return "sleep" / parens(self.format(node.arguments))

    @format.register
    def _(self, node: TalonVariable) -> Doc:
        self.assert_only_comments(node.children)
        return self.format(node.variable_name)

    ###########################################################################
    # Format: Numbers
    ###########################################################################

    @format.register
    def _(self, node: TalonFloat) -> Doc:
        return Text.words(node.text.strip(), collapse_whitespace=True)

    @format.register
    def _(self, node: TalonInteger) -> Doc:
        return Text.words(node.text.strip(), collapse_whitespace=True)

    @format.register
    def _(self, node: TalonNumber) -> Doc:
        return self.format(self.get_node(node.children))

    ###########################################################################
    # Format: Strings
    ###########################################################################

    @format.register
    def _(self, node: TalonImplicitString) -> Doc:
        return Text.words(node.text.strip(), collapse_whitespace=True)

    @format.register
    def _(self, node: TalonInterpolation) -> Doc:
        return self.format(self.get_node(node.children))

    @format.register
    def _(self, node: TalonString) -> Doc:
        return double_quote(self.format_children(node.children))

    @format.register
    def _(self, node: TalonStringContent) -> Doc:
        return Text.words(node.text)

    @format.register
    def _(self, node: TalonStringEscapeSequence) -> Doc:
        return Text.words(node.text)

    ###########################################################################
    # Format: Rules
    ###########################################################################

    @format.register
    def _(self, node: TalonCapture) -> Doc:
        self.assert_only_comments(node.children)
        return angles(self.format(node.capture_name))

    @format.register
    def _(self, node: TalonChoice) -> Doc:
        children = self.format_children(node.children)
        operator = Space / "|" / Space
        return operator.join(children)

    @format.register
    def _(self, node: TalonEndAnchor) -> Doc:
        return Text("$")

    @format.register
    def _(self, node: TalonList) -> Doc:
        self.assert_only_comments(node.children)
        return braces(self.format(node.list_name))

    @format.register
    def _(self, node: TalonOptional) -> Doc:
        child = self.get_node(node.children)
        return brackets(self.format(child))

    @format.register
    def _(self, node: TalonParenthesizedRule) -> Doc:
        child = self.get_node(node.children)
        return parens(self.format(child))

    @format.register
    def _(self, node: TalonRepeat) -> Doc:
        child = self.get_node(node.children)
        return self.format(child) / "*"

    @format.register
    def _(self, node: TalonRepeat1) -> Doc:
        return self.format(self.get_node(node.children)) / "+"

    @format.register
    def _(self, node: TalonRule) -> Doc:
        return cat(self.format_children(node.children))

    @format.register
    def _(self, node: TalonSeq) -> Doc:
        return Space.join(self.format_children(node.children))

    @format.register
    def _(self, node: TalonStartAnchor) -> Doc:
        return Text("^")

    @format.register
    def _(self, node: TalonWord) -> Doc:
        return Text.words(node.text)

    ###########################################################################
    # Format: Comments
    ###########################################################################

    @format.register
    def _(self, node: TalonComment) -> Doc:
        if self.format_comments:
            # TODO: format blocks of comments so we can:
            #       1. decrease indentation consistently;
            #       2. reflow text
            return (
                "#"
                // Text.words(node.text.lstrip("#"), collapse_whitespace=False)
                / Line
            )
        else:
            return Text.words(node.text, collapse_whitespace=False) / Line

    @format.register
    def _(self, node: TalonDocstring) -> Doc:
        if self.format_comments:
            # TODO: format blocks of comments so we can:
            #       1. decrease indentation consistently;
            #       2. reflow text
            return (
                "###"
                // Text.words(node.text.lstrip("#"), collapse_whitespace=False)
                / Line
            )
        else:
            return Text.words(node.text, collapse_whitespace=False) / Line

    # Used to buffer comments encountered inline, e.g., inside a binary operator
    _comment_buffer: list[TalonComment] = dataclasses.field(
        default_factory=list, init=False
    )

    def store_comments_with_type(
        self,
        children: Iterable[Union[TalonComment, NodeVar]],
        *,
        node_type: type[NodeVar],
    ) -> Iterator[NodeVar]:
        """
        Store all the comments in the iterable, yield the rest.
        """
        for child in children:
            if isinstance(child, TalonComment):
                self._comment_buffer.append(child)
            elif isinstance(child, node_type):
                yield child
            else:
                raise TypeError(type(child))

    def get_comments(self) -> Iterator[TalonComment]:
        """
        Get the buffered comments. Clear the buffer.
        """
        try:
            yield from self._comment_buffer
        finally:
            self._comment_buffer.clear()

    def get_formatted_comments(self) -> Iterator[Doc]:
        """
        Get the buffered comments, formatted. Clear the buffer.
        """
        yield from map(self.format, self.get_comments())

    def assert_only_comments(self, children: Iterable[TalonComment]) -> None:
        """
        Assert that all the nodes in the iterable are comments.
        """
        rest = tuple(self.store_comments_with_type(children, node_type=TalonComment))
        assert (
            len(rest) == 0
        ), f"There should be no non-comment nodes, found {tuple(node.type_name for node in rest)}:\n{rest}"

    def get_node(self, children: Iterable[Node]) -> Node:
        """
        Get the single node that is not a comment. Store all the comments.
        """
        return self.get_node_with_type(children, node_type=Node)

    def get_node_with_type(
        self,
        children: Iterable[Union[NodeVar, TalonComment]],
        *,
        node_type: type[NodeVar],
    ) -> NodeVar:
        """
        Get the single node that is not a comment, but has type NodeVar. Store all the comments.
        """
        rest = tuple(self.store_comments_with_type(children, node_type=node_type))
        assert (
            len(rest) == 1
        ), f"There should be only one non-comment child, found {tuple(node.type_name for node in rest)}:\n{rest}"
        return next(iter(rest))
